
/*______________________________________________________________________________
tile engine
______________________________________________________________________________*/

/*

*/
CREATE TABLE grp(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    coll INTEGER NOT NULL,
    rowe INTEGER NOT NULL,
    name TEXT NOT NULL
);

/*

*/
CREATE TABLE tile(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    coll INTEGER NOT NULL,
    rowe INTEGER NOT NULL,
    symbol TEXT NOT NULL,
    id_grp INTEGER NOT NULL,
    FOREIGN KEY(id_grp) REFERENCES grp(id)
);

/*

*/
CREATE VIEW tileGrp AS
SELECT tile.id, 
       tile.coll+grp.coll AS coll, 
       tile.rowe+grp.rowe AS rowe,
       tile.symbol,
       tile.id_grp,
       tile.name
FROM tile
INNER JOIN grp
ON grp.id = tile.id_grp;

/*

*/
CREATE VIEW neighbor AS
SELECT 0 AS coll,1 AS rowe
UNION ALL
SELECT 0,-1
UNION ALL
SELECT 1,0
UNION ALL
SELECT -1,0;

/*

*/
CREATE VIEW tileNeighbor AS
SELECT tileGrp.id AS id,
       tileGrp.coll AS coll,
       tileGrp.rowe AS rowe,
       tileGrp.symbol AS symbol,
       tileGrp.name AS name,
       tileGrp.id_grp AS id_grp,
       tileGrp.coll + neighbor.coll AS collAdj,
       tileGrp.rowe + neighbor.rowe AS roweAdj,
       neighbor.name AS neighbor
FROM tileGrp
JOIN neighbor;

/*

*/
CREATE VIEW tileAdjacent AS
SELECT tileNeighbor.id,
       tileNeighbor.coll,
       tileNeighbor.rowe,
       tileNeighbor.symbol,
       tileNeighbor.id_grp,
       tileNeighbor.collAdj,
       tileNeighbor.roweAdj,
       tileGrp.symbol AS symbAdj,
       tileGrp.id_grp AS id_grpAdj,
       tileGrp.name AS nameAdj
FROM tileNeighbor
INNER JOIN tileGrp 
ON tileGrp.coll = tileNeighbor.collAdj 
AND tileGrp.rowe = tileNeighbor.roweAdj;

/*______________________________________________________________________________
fsm engine
______________________________________________________________________________*/

/*

*/
CREATE TABLE tag(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL UNIQUE
);

CREATE TABLE tagging(
    id_tag INTEGER NOT NULL,
    id_grp INTEGER NOT NULL,
    FOREIGN KEY (id_tag) REFERENCES tag(id),
    FOREIGN KEY (id_grp) REFERENCES grp(id)
);

CREATE VIEW grpTag AS
SELECT tileGrp.id AS id,
       tileGrp.id_grp AS id_grp,
       tileGrp.name AS name,
       tileGrp.coll AS coll,
       tileGrp.rowe AS rowe,
       tileGrp.symbol AS symbol,
       tag.name AS tag
FROM tagging
INNER JOIN tileGrp ON tagging.id_grp = tileGrp.id
INNER JOIN tag ON tagging.id_tag = tag.id;
