
import defensive

log = defensive.Logging('fsm')

@defensive.arg(str,(list,tuple),(str,type(None)))
@defensive.pre(lambda name,trans,default : False not in [type(trigger)==type(lambda a:True) and type(state)==str for trigger,state in trans])
@defensive.pre(lambda name,trans,default : name != '')
@defensive.pre(lambda name,trans,default : default != '' or default==None)
def state(stateName, transitionList=[],default=None):
	transition = tuple(transitionList)
	@log.debug
	def closure():
		for (trigger,nextState) in transition:
			if trigger():
				return nextState
		if not default:
			return stateName
		return default
	closure.__name__=stateName
	return closure

@log.debug
def _stateFromJson(data={},triggerSuite={}):
	"""
	json:
	{
		"name":string,
		"transition":{"triggerPredicateName":"nextStateName"},
		"default":"defaultStateTransition"
	}
	will return state closure
	"""
	stateName = data['name']
	transitionList = [(triggerSuite[triggerName],nextState) for triggerName,nextState in data['transition'].items()]
	default = data['default'] if 'default' in data.keys() else stateName
	return state(stateName,transitionList,default)

@log.debug
def fromJson(data=[],triggerSuite={}):
	"""
	json:[{...see _stateFromJson ...},{...}]
	will return dictionnary by state name containing all state
	"""
	stateMachine = {}
	for subData in data:
		state = _stateFromJson(subData,triggerSuite)
		stateMachine[subData['name']]=state
	return stateMachine

if __name__ == '__main__':
	import logging
	log.logger.setLevel(logging.DEBUG)
	log.logger.addHandler(logging.StreamHandler())
	test= defensive.Testing()

	machine = [
		{	'name':'A',	'transition':{'everyFiveSec':'B','everySevenSec':'C'}},
		{	'name':'B',	'transition':{},'default':'A'},
		{	'name':'C',	'transition':{'everyThreeSec':'D'}},
		{	'name':'D',	'transition':{}	}
	]
	time = 1
	def everyThreeSec():
		return time%3==0
	def everyFiveSec():
		return time%5==0
	def everySevenSec():
		return time%7==0
	triggersuite={
		'everyThreeSec':everyThreeSec,
		'everyFiveSec':everyFiveSec,
		'everySevenSec':everySevenSec
	}
	fsm = fromJson(machine,triggersuite)
	
	@test.eq('B')
	def transitionA5sec():
		global time
		time = 5
		return fsm['A']()

	@test.eq('A')
	def transitionA3sec():
		global time
		time = 3
		return fsm['A']()

	@test.eq('C')
	def transitionA7sec():
		global time
		time = 7
		return fsm['A']()

	@test.eq('C')
	def transitionA15sec():
		global time
		time = 7
		return fsm['A']()

	@test.eq('A')
	def transitionB():
		return fsm['B']()

	@test.eq('D')
	def transitionC3sec():
		global time
		time = 3
		return fsm['C']()

	@test.eq('C')
	def transitionC5sec():
		global time
		time = 5
		return fsm['C']()
	
	@test.exception(Exception,{'notAFunction':'astring'})
	@test.exception(Exception,{lambda a:True:5})
	def invalidTransition(trans):
		state('whatever',trans)
	
	test.runTest()
