import sqlite3

dbinit = """
PRAGMA foreign_keys = ON;


DROP VIEW IF EXISTS nodeDenormalised;
DROP VIEW IF EXISTS identDenormalised;
DROP VIEW IF EXISTS dataDenormalised;
DROP TABLE IF EXISTS constant;
DROP TABLE IF EXISTS node;
DROP TABLE IF EXISTS data;
DROP TABLE IF EXISTS type;
DROP TABLE IF EXISTS ident;

CREATE TABLE constant(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL,
    value TEXT NOT NULL
);

CREATE TABLE ident(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL UNIQUE,
    admin_id INTEGER NULL,
    FOREIGN KEY(admin_id) REFERENCES ident(id)
);
INSERT INTO ident (name) VALUES('boss');

CREATE TABLE type(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL UNIQUE
);
INSERT INTO type (name) VALUES ('dir');
INSERT INTO constant (name,value) VALUES('dir_type_id', last_insert_rowid());
INSERT INTO type (name) VALUES ('link');
INSERT INTO constant (name,value) VALUES('link_type_id', last_insert_rowid());
INSERT INTO type (name) VALUES ('text');
INSERT INTO constant (name,value) VALUES('text_type_id', last_insert_rowid());
INSERT INTO type (name) VALUES ('exec');
INSERT INTO constant (name,value) VALUES('exec_type_id', last_insert_rowid());

CREATE TABLE data(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    content TEXT NOT NULL
);

CREATE TABLE node(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL,
    id_data INTEGER NULL,
    id_type INTEGER NOT NULL,
    id_parent INTEGER NULL,
    id_ident_read INTEGER NOT NULL,
    id_ident_write INTEGER NOT NULL,
    FOREIGN KEY(id_data) REFERENCES data(id),
    FOREIGN KEY(id_type) REFERENCES type(id),
    FOREIGN KEY(id_parent) REFERENCES node(id),
    FOREIGN KEY(id_ident_read) REFERENCES ident(id),
    FOREIGN KEY(id_ident_write) REFERENCES ident(id)
);
INSERT INTO node (name,id_type,id_ident_read,id_ident_write) 
SELECT 'root', type.id, ident.id, ident.id FROM type 
INNER JOIN ident ON ident.name='boss' 
WHERE type.name = 'dir' ;
INSERT INTO constant (name,value) VALUES('rootid', last_insert_rowid());

CREATE VIEW nodeDenormalised AS 
SELECT node.id,
       parent.id AS id_parent, 
       node.name, 
       typetbl.name AS type, 
       canRead.name AS canRead, 
       canWrite.name as canWrite, 
       parent.name AS parent
FROM node 
INNER JOIN type AS typetbl ON node.id_type = typetbl.id 
INNER JOIN ident AS canRead ON node.id_ident_read = canRead.id 
INNER JOIN ident AS canWrite ON node.id_ident_write = canWrite.id 
LEFT JOIN node AS parent ON node.id_parent = parent.id ;

CREATE VIEW identDenormalised AS 
SELECT ident.id AS id,
       ident.name, 
       admin.id AS idadmin,
       admin.name AS admin
FROM ident 
LEFT JOIN ident AS admin ON ident.admin_id = admin.id ;

CREATE VIEW dataDenormalised AS 
SELECT node.id,
       node.name,
       data.content AS data, 
       canRead.name AS canRead, 
       canWrite.name as canWrite
FROM node 
INNER JOIN data ON id_data = data.id 
INNER JOIN ident AS canRead ON id_ident_read = canRead.id 
INNER JOIN ident AS canWrite ON id_ident_write = canWrite.id ;
"""

testinit = """
INSERT INTO node (name,id_type,id_ident_read,id_ident_write,id_parent) 
SELECT 'bin', type.id, ident.id, ident.id, parent.id
FROM type 
INNER JOIN ident ON ident.name='boss' 
INNER JOIN node AS parent ON parent.name='root' AND parent.id_parent IS NULL 
WHERE type.name = 'dir' ;

INSERT INTO node (name,id_type,id_ident_read,id_ident_write,id_parent) 
SELECT 'dev', type.id, ident.id, ident.id, parent.id
FROM type 
INNER JOIN ident ON ident.name='boss' 
INNER JOIN node AS parent ON parent.name='root' AND parent.id_parent IS NULL 
WHERE type.name = 'dir' ;

INSERT INTO node (name,id_type,id_ident_read,id_ident_write,id_parent) 
SELECT 'etc', type.id, ident.id, ident.id, parent.id
FROM type 
INNER JOIN ident ON ident.name='boss' 
INNER JOIN node AS parent ON parent.name='root' AND parent.id_parent IS NULL 
WHERE type.name = 'dir' ;

INSERT INTO node (name,id_type,id_ident_read,id_ident_write,id_parent) 
SELECT 'mnt', type.id, ident.id, ident.id, parent.id
FROM type 
INNER JOIN ident ON ident.name='boss' 
INNER JOIN node AS parent ON parent.name='root' AND parent.id_parent IS NULL 
WHERE type.name = 'dir' ;

INSERT INTO node (name,id_type,id_ident_read,id_ident_write,id_parent) 
SELECT 'usr', type.id, ident.id, ident.id, parent.id
FROM type 
INNER JOIN ident ON ident.name='boss' 
INNER JOIN node AS parent ON parent.name='root' AND parent.id_parent IS NULL 
WHERE type.name = 'dir' ;


INSERT INTO node (name,id_type,id_ident_read,id_ident_write,id_parent) 
SELECT 'welcome', type.id, ident.id, ident.id, parent.id
FROM type 
INNER JOIN ident ON ident.name='boss' 
INNER JOIN node AS parent ON parent.name='usr' 
WHERE type.name = 'text' ;
"""
#hope to transform this into a shell script (it will exist)

sqlite3.enable_callback_tracebacks(True)
class MainframeDao(object):
    def __init__(self, name='mainframe.db'):
        self.conn = sqlite3.connect(name)
        cur = self.conn.cursor()
        cur.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='type';")
        if not cur.fetchone():
            cur.executescript(dbinit);
            cur.executescript(testinit);
            self.conn.commit()
    
    def _buildWhere(self,where):
        sql = ""
        for col,val in where.items():
            sql += "AND "+col+"=:"+col+" "
        return sql
    
    def getNode(self,where={}):
        sql = """
        SELECT * FROM nodeDenormalised 
        WHERE 1=1 
        """
        sql += self._buildWhere(where)
        cur = self.conn.cursor()
        cur.execute(sql,where)
        for row in cur.fetchall():
            yield row
    
    def getIdent(self,where={}):
        sql = """
        SELECT * FROM identDenormalised 
        WHERE 1=1 
        """
        for col,val in where.items():
            sql += "AND "+col+"=:"+col+" "
        cur = self.conn.cursor()
        cur.execute(sql,where)
        for row in cur.fetchall():
            yield row
    
    def getData(self,where={}):
        sql = """
        SELECT * FROM dataDenormalised 
        WHERE 1=1 
        """
        for col,val in where.items():
            sql += "AND "+col+"=:"+col+" "
        cur = self.conn.cursor()
        cur.execute(sql,where)
        for row in cur.fetchall():
            yield row
    
    def getConstant(self,name):
        cur = self.conn.cursor()
        cur.execute("SELECT value FROM constant WHERE name=?",(name,))
        return cur.fetchone()[0]

    def addNode(self,name,typeid,canReadId,canWriteId,parentId):
        sql = """
        INSERT INTO node (name,id_type,id_ident_read,id_ident_write,id_parent)
        VALUES (?,?,?,?,?)
        """
        cur = self.conn.cursor()
        cur.execute(sql,(name,typeid,canReadId,canWriteId,parentId))
        cur.execute('SELECT last_insert_rowid()')
        return cur.fetchone()[0]
    
    def addIdent(self,name,admin):
        sql = """
        INSERT INTO ident (name,admin_id)
        VALUES (?,?)
        """
        cur = self.conn.cursor()
        cur.execute(sql,(name,admin))
        cur.execute('SELECT last_insert_rowid()')
        return cur.fetchone()[0]
    
    def readData(self,nodeid):
        sql = """
        SELECT data FROM dataDenormalised 
        WHERE id = ?
        """
        cur = self.conn.cursor()
        cur.execute(sql,(nodeid,))
        data = cur.fetchone()
        print data
        if data:
            return data
        return ''
    
    def writeData(self,data,nodeid):
        cur = self.conn.cursor()
        sql = """
        INSERT INTO data (content) VALUES(?)
        """
        cur.execute(sql,(data,))
        cur.execute('SELECT last_insert_rowid()')
        dataid = cur.fetchone()[0]
        sql = """
        UPDATE node SET id_data=? WHERE id=? 
        """
        cur.execute(sql,(dataid,nodeid))
        sql = """
        DELETE FROM data 
        WHERE id NOT IN (SELECT id_data FROM node)
        """
        cur = self.conn.cursor()
        cur.execute(sql)
        cur.execute('SELECT last_insert_rowid()')
        return cur.fetchone()[0]
        
    
    def changeWriter(self,nodeid,writerid):
        sql = """
        UPDATE node SET id_ident_write=? WHERE id=? 
        """
        cur = self.conn.cursor()
        cur.execute(sql,(writerid,nodeid))
    
    def changeReader(self,nodeid,readerid):
        sql = """
        UPDATE node SET id_ident_read=? WHERE id=? 
        """
        cur = self.conn.cursor()
        cur.execute(sql,(readerid,nodeid))
    
class Terminal(object):
    def __init__(self,mainframeDao,nodeid,ident=[]):
        self.dao=mainframeDao
        self.nodeid = nodeid
        self.ident = ident

    def ls(self):
        for node in self.dao.getNode({'id_parent':self.nodeid}):
            if node[4] in self.ident:
                yield node[2]

    def pwd(self):
        for node in self.dao.getNode({'id':self.nodeid}):
            return node[2:7]
    
    def cd(self,name='.'):
        if name=='.':
            return self
        if name=='..':
            node = self.pwd()
            if (node and node[4] and node[3] in self.ident):
                return Terminal(self.dao, node[1], self.ident)
            raise Exception(name+" not exist or not allowed")
        for node in self.dao.getNode({'id_parent':self.nodeid}):
            if node[2]==name:
                return Terminal(self.dao, node[0], self.ident)
        raise Exception(name+" not exist")

    def touch(self,name):
        if self.pwd()[3] in self.ident and name not in self.ls():
            typeid = self.dao.getConstant('text_type_id')
            for ident in self.dao.getIdent({'name':self.ident[0]}):
                identid = ident[0]
                self.dao.addNode(name,typeid,identid,identid,self.nodeid)
                return True
        raise Exception(name+" not allowed")     
    
    def read(self):
        if self.pwd()[2] in self.ident:
            return self.dao.readData(self.nodeid)
        raise Exception(name+" not allowed")
    
    def write(self,data):
        if self.pwd()[3] in self.ident:
            return self.dao.writeData(data,self.nodeid) != None
        raise Exception(name+" not allowed")
    
    def chmod(self,reader,writer):
        if self.pwd()[3] in self.ident:
            for readerNode in self.dao.getIdent({'name':reader}):
                readerid = readerNode[0]
                self.dao.changeReader(self.nodeid,readerid)
            writerid = readerid
            if writer!=reader:
                for writerNode in self.dao.getIdent({'name':writer}):
                    writerid = writerNode[0]
            self.dao.changeWriter(self.nodeid,writerid)
            return True
        raise Exception(name+" not allowed")
    
    def addUser(self,user):
        for ident in self.dao.getIdent({'name':self.ident[0]}):
            return self.dao.addIdent(user,ident[0])
    
    def users(self):
        for ident in self.dao.getIdent():
            yield ident
        
class Mainframe(object):
    def __init__(self, name='mainframe.db'):
        self.dao = MainframeDao(name)

    def connect(self,ident=[]):
        rootid=self.dao.getConstant('rootid')
        return Terminal(self.dao,rootid,ident)

        
if __name__ == '__main__':
    mainframe = Mainframe(':memory:')
    term = mainframe.connect(['boss'])

    print '=>pwd' , term.pwd()
    print '=>ls'
    for i in term.ls():
        print i
        
    print '=>cd usr'
    print '=>pwd' , term.cd('usr').pwd()
    print '=>ls'
    for i in term.cd('usr').ls():
        print i
        
    print '=>cd ..'
    print term.pwd()
    print '=>pwd' , term.cd('usr').cd('..').pwd()
    print '=>ls'
    for i in term.cd('usr').cd('..').ls():
        print i
        
    print '=>cd .'
    print '=>pwd' , term.cd('.').pwd()
    print '=>ls'
    for i in term.ls():
        print i
    
    try:
        print '=>cd ..'
        term.cd('..')
    except:
        print '=>ok'
    
    try:
        print '=>cd rien'
        term.cd('rien')
    except:
        print '=>ok'
    
    print '=>touch file'
    term.cd('usr').touch('file')
    print '=>write test into file'
    print term.cd('usr').cd('file').write('test')
    print '=>read file'
    print term.cd('usr').cd('file').read()
    
    print '=>addUser'
    term.addUser('monsieurs')
    for user in term.users():
        print user
    print '=>before chmod'
    for i in term.cd('usr').ls():
        print i
    print '=>after chmod'
    term.cd('usr').cd('file').chmod('monsieurs','monsieurs')
    for i in term.cd('usr').ls():
        print i
        
